@@ -4,6 +4,8 @@
 # Distributed under the terms of the Modified BSD License.
 
 from collections import namedtuple
+from itertools import chain
+from zmq import POLLOUT, POLLIN
 
 try:
     from tornado.concurrent import Future
@@ -28,7 +30,7 @@ import zmq as _zmq
 from zmq.eventloop.ioloop import IOLoop
 
 
-_FutureEvent = namedtuple('_FutureEvent', ('future', 'kind', 'args', 'msg'))
+_FutureEvent = namedtuple('_FutureEvent', ('future', 'kind', 'kwargs', 'msg'))
 
 # mixins for tornado/asyncio compatibility
 
@@ -42,6 +44,7 @@ class _AsyncTornado(object):
 
 class _AsyncPoller(_zmq.Poller):
     """Poller that returns a Future on poll, instead of blocking."""
+
     def poll(self, timeout=-1):
         """Return a Future for a poll event"""
         future = self._Future()
@@ -57,13 +60,30 @@ class _AsyncPoller(_zmq.Poller):
         loop = self._default_loop()
         
         # register Future to be called as soon as any event is available on any socket
-        # only support polling on zmq sockets, for now
         watcher = self._Future()
+        
+        # watch raw sockets:
+        raw_sockets = []
+        def wake_raw(*args):
+            if not watcher.done():
+                watcher.set_result(None)
+
+        watcher.add_done_callback(lambda f: self._unwatch_raw_sockets(loop, *raw_sockets))
+
         for socket, mask in self.sockets:
-            if mask & _zmq.POLLIN:
-                socket._add_recv_event('poll', future=watcher)
-            if mask & _zmq.POLLOUT:
-                socket._add_send_event('poll', future=watcher)
+            if isinstance(socket, _AsyncSocket):
+                if mask & _zmq.POLLIN:
+                    socket._add_recv_event('poll', future=watcher)
+                if mask & _zmq.POLLOUT:
+                    socket._add_send_event('poll', future=watcher)
+            else:
+                raw_sockets.append(socket)
+                evt = 0
+                if mask & _zmq.POLLIN:
+                    evt |= self._READ
+                if mask & _zmq.POLLOUT:
+                    evt |= self._WRITE
+                self._watch_raw_socket(loop, socket, evt, wake_raw)
         
         def on_poll_ready(f):
             if future.done():
@@ -79,7 +99,7 @@ class _AsyncPoller(_zmq.Poller):
                     future.set_result(result)
         watcher.add_done_callback(on_poll_ready)
         
-        if timeout > 0:
+        if timeout is not None and timeout > 0:
             # schedule cancel to fire on poll timeout, if any
             def trigger_timeout():
                 if not watcher.done():
@@ -95,16 +115,24 @@ class _AsyncPoller(_zmq.Poller):
                 else:
                     loop.remove_timeout(timeout_handle)
             future.add_done_callback(cancel_timeout)
-        
+
         def cancel_watcher(f):
             if not watcher.done():
                 watcher.cancel()
         future.add_done_callback(cancel_watcher)
-            
+
         return future
 
 class Poller(_AsyncTornado, _AsyncPoller):
-    pass
+    def _watch_raw_socket(self, loop, socket, evt, f):
+        """Schedule callback for a raw socket"""
+        loop.add_handler(socket, lambda *args: f(), evt)
+
+    def _unwatch_raw_sockets(self, loop, *sockets):
+        """Unschedule callback for a raw socket"""
+        for socket in sockets:
+            loop.remove_handler(socket)
+
 
 class _AsyncSocket(_zmq.Socket):
     
@@ -114,7 +142,7 @@ class _AsyncSocket(_zmq.Socket):
     _shadow_sock = None
     _poller_class = Poller
     io_loop = None
-    
+
     def __init__(self, context, socket_type, io_loop=None):
         super(_AsyncSocket, self).__init__(context, socket_type)
         self.io_loop = io_loop or self._default_loop()
@@ -123,7 +151,16 @@ class _AsyncSocket(_zmq.Socket):
         self._state = 0
         self._shadow_sock = _zmq.Socket.shadow(self.underlying)
         self._init_io_state()
-    
+
+    def close(self, linger=None):
+        if not self.closed:
+            for event in chain(self._recv_futures, self._send_futures):
+                if not event.future.done():
+                    event.future.cancel()
+            self._clear_io_state()
+        super(_AsyncSocket, self).close(linger=linger)
+    close.__doc__ = _zmq.Socket.close.__doc__
+
     def recv_multipart(self, flags=0, copy=True, track=False):
         """Receive a complete multipart zmq message.
         
@@ -150,7 +187,7 @@ class _AsyncSocket(_zmq.Socket):
         Returns a Future that resolves when sending is complete.
         """
         return self._add_send_event('send_multipart', msg=msg,
-            args=dict(flags=flags, copy=copy, track=track),
+            kwargs=dict(flags=flags, copy=copy, track=track),
         )
     
     def send(self, msg, flags=0, copy=True, track=False):
@@ -161,9 +198,38 @@ class _AsyncSocket(_zmq.Socket):
         Recommend using send_multipart instead.
         """
         return self._add_send_event('send', msg=msg,
-            args=dict(flags=flags, copy=copy, track=track),
+            kwargs=dict(flags=flags, copy=copy, track=track),
         )
-    
+
+    def _deserialize(self, recvd, load):
+        """Deserialize with Futures"""
+        f = self._Future()
+        def _chain(_):
+            """Chain result through serialization to recvd"""
+            if f.done():
+                return
+            if recvd.exception():
+                f.set_exception(recvd.exception())
+            else:
+                buf = recvd.result()
+                try:
+                    loaded = load(buf)
+                except Exception as e:
+                    f.set_exception(e)
+                else:
+                    f.set_result(loaded)
+        recvd.add_done_callback(_chain)
+
+        def _chain_cancel(_):
+            """Chain cancellation from f to recvd"""
+            if recvd.done():
+                return
+            if f.cancelled():
+                recvd.cancel()
+        f.add_done_callback(_chain_cancel)
+
+        return f
+
     def poll(self, timeout=None, flags=_zmq.POLLIN):
         """poll the socket for events
         
@@ -182,34 +248,116 @@ class _AsyncSocket(_zmq.Socket):
             if future.done():
                 return
             if f.exception():
-                future.set_exception(f.exeception())
+                future.set_exception(f.exception())
             else:
                 evts = dict(f.result())
                 future.set_result(evts.get(self, 0))
-        
+
         f.add_done_callback(unwrap_result)
         return future
 
-    def _add_recv_event(self, kind, args=None, future=None):
+    def _add_timeout(self, future, timeout):
+        """Add a timeout for a send or recv Future"""
+        def future_timeout():
+            if future.done():
+                # future already resolved, do nothing
+                return
+
+            # pop the entry from _recv_futures
+            for f_idx, (f, kind, kwargs, _) in enumerate(self._recv_futures):
+                if f == future:
+                    self._recv_futures.pop(f_idx)
+                    break
+
+            # pop the entry from _send_futures
+            for f_idx, (f, kind, kwargs, _) in enumerate(self._send_futures):
+                if f == future:
+                    self._send_futures.pop(f_idx)
+                    break
+
+            # raise EAGAIN
+            future.set_exception(_zmq.Again())
+        self._call_later(timeout, future_timeout)
+
+    def _call_later(self, delay, callback):
+        """Schedule a function to be called later
+
+        Override for different IOLoop implementations
+
+        Tornado and asyncio happen to both have ioloop.call_later
+        with the same signature.
+        """
+        self.io_loop.call_later(delay, callback)
+
+    def _add_recv_event(self, kind, kwargs=None, future=None):
         """Add a recv event, returning the corresponding Future"""
         f = future or self._Future()
+        if kind.startswith('recv') and kwargs.get('flags', 0) & _zmq.DONTWAIT:
+            # short-circuit non-blocking calls
+            recv = getattr(self._shadow_sock, kind)
+            try:
+                r = recv(**kwargs)
+            except Exception as e:
+                f.set_exception(e)
+            else:
+                f.set_result(r)
+            return f
+
+        # we add it to the list of futures before we add the timeout as the
+        # timeout will remove the future from recv_futures to avoid leaks
         self._recv_futures.append(
-            _FutureEvent(f, kind, args, msg=None)
+            _FutureEvent(f, kind, kwargs, msg=None)
         )
-        self._add_io_state(self._READ)
+
+        if hasattr(_zmq, 'RCVTIMEO'):
+            timeout_ms = self._shadow_sock.rcvtimeo
+            if timeout_ms >= 0:
+                self._add_timeout(f, timeout_ms * 1e-3)
+
+        if self.events & POLLIN:
+            # recv immediately, if we can
+            self._handle_recv()
+        if self._recv_futures:
+            self._add_io_state(self._READ)
         return f
     
-    def _add_send_event(self, kind, msg=None, args=None, future=None):
-        """Add a recv event, returning the corresponding Future"""
+    def _add_send_event(self, kind, msg=None, kwargs=None, future=None):
+        """Add a send event, returning the corresponding Future"""
         f = future or self._Future()
+        if kind.startswith('send') and kwargs.get('flags', 0) & _zmq.DONTWAIT:
+            # short-circuit non-blocking calls
+            send = getattr(self._shadow_sock, kind)
+            try:
+                r = send(msg, **kwargs)
+            except Exception as e:
+                f.set_exception(e)
+            else:
+                f.set_result(r)
+            return f
+
+        # we add it to the list of futures before we add the timeout as the
+        # timeout will remove the future from recv_futures to avoid leaks
         self._send_futures.append(
-            _FutureEvent(f, kind, args=args, msg=msg)
+            _FutureEvent(f, kind, kwargs=kwargs, msg=msg)
         )
-        self._add_io_state(self._WRITE)
+
+        if hasattr(_zmq, 'SNDTIMEO'):
+            timeout_ms = self._shadow_sock.sndtimeo
+            if timeout_ms >= 0:
+                self._add_timeout(f, timeout_ms * 1e-3)
+
+        if self.events & POLLOUT:
+            # send immediately if we can
+            self._handle_send()
+        if self._send_futures:
+            self._add_io_state(self._WRITE)
         return f
     
     def _handle_recv(self):
         """Handle recv events"""
+        if not self._shadow_sock.events & POLLIN:
+            # event triggered, but state may have been changed between trigger and callback
+            return
         f = None
         while self._recv_futures:
             f, kind, kwargs, _ = self._recv_futures.pop(0)
@@ -245,6 +393,9 @@ class _AsyncSocket(_zmq.Socket):
             f.set_result(result)
     
     def _handle_send(self):
+        if not self._shadow_sock.events & POLLOUT:
+            # event triggered, but state may have been changed between trigger and callback
+            return
         f = None
         while self._send_futures:
             f, kind, kwargs, msg = self._send_futures.pop(0)
@@ -303,11 +454,19 @@ class _AsyncSocket(_zmq.Socket):
         """Update IOLoop handler with state."""
         self._state = state
         self.io_loop.update_handler(self, state)
-    
+
     def _init_io_state(self):
         """initialize the ioloop event handler"""
         self.io_loop.add_handler(self, self._handle_events, self._state)
 
+    def _clear_io_state(self):
+        """unregister the ioloop event handler
+        
+        called once during close
+        """
+        self.io_loop.remove_handler(self)
+
+
 class Socket(_AsyncTornado, _AsyncSocket):
     pass
 
